@@ -5,12 +5,12 @@ to be available locally."""
 import re
 import os
 import csv
+import gzip
 import json
 import logging
 import requests
 import itertools
 import indra
-from indra.util import write_unicode_csv
 from indra.databases import hgnc_client, uniprot_client, chebi_client, \
     go_client, mesh_client, doid_client
 from indra.statements.resources import amino_acids
@@ -55,7 +55,7 @@ def generate_hgnc_terms():
             new_id = m.groups()[0]
             new_name = id_name_map[new_id]
             term_args = (normalize(name), name, db, new_id,
-                         new_name, 'previous', 'hgnc', organism)
+                         new_name, 'former_name', 'hgnc', organism)
             all_term_args[term_args] = None
             # NOTE: consider adding withdrawn synonyms e.g.,
             # symbol withdrawn, see pex1     symbol withdrawn, see PEX1
@@ -86,7 +86,7 @@ def generate_hgnc_terms():
             prev_symbols = row['Previous symbols'].split(', ')
             for prev_symbol in prev_symbols:
                 term_args = (normalize(prev_symbol), prev_symbol, db, id, name,
-                             'previous', 'hgnc', organism)
+                             'former_name', 'hgnc', organism)
                 all_term_args[term_args] = None
 
     terms = [Term(*args) for args in all_term_args.keys()]
@@ -164,6 +164,7 @@ def generate_mesh_terms(ignore_mappings=False):
                        'mesh_supp_id_label_mappings.tsv']
     terms = []
     for fname in mesh_name_files:
+        logger.info('Loading %s' % fname)
         mesh_names_file = os.path.join(indra_resources, fname)
         for row in read_csv(mesh_names_file, header=False, delimiter='\t'):
             db_id = row[0]
@@ -178,14 +179,18 @@ def generate_mesh_terms(ignore_mappings=False):
                 status = 'name'
                 name = text_name
             term = Term(normalize(text_name), text_name, db, db_id, name,
-                        status, 'mesh')
+                        status, 'mesh',
+                        source_db='MESH' if db != 'MESH' else None,
+                        source_id=row[0] if db != 'MESH' else None)
             terms.append(term)
             synonyms = row[2]
             if row[2]:
                 synonyms = synonyms.split('|')
                 for synonym in synonyms:
                     term = Term(normalize(synonym), synonym, db, db_id, name,
-                                'synonym', 'mesh')
+                                'synonym', 'mesh',
+                                source_db='MESH' if db != 'MESH' else None,
+                                source_id=row[0] if db != 'MESH' else None)
                     terms.append(term)
         logger.info('Loaded %d terms' % len(terms))
     return terms
@@ -206,6 +211,13 @@ def generate_go_terms():
         # Next look at synonyms, sometimes those match the name so we
         # deduplicate
         for synonym in set(entry.get('synonyms', [])) - {name}:
+            # GO includes around 40k synonyms for terms that represent
+            # activity out of which around 5k are actually
+            # synonyms for entities. One example is "EGFR" as a synonym
+            # for "epidermal growth factor-activated receptor activity".
+            # We skip these according to the following logic.
+            if 'activity' in name and 'activity' not in synonym:
+                continue
             term = Term(normalize(synonym), synonym, 'GO', go_id, name,
                         'synonym', 'go')
             terms.append(term)
@@ -223,11 +235,11 @@ def generate_famplex_terms(ignore_mappings=False):
         groundings = {k: v for k, v in zip(row[1::2], row[2::2]) if (k and v)}
         if 'FPLX' in groundings:
             id = groundings['FPLX']
-            term = Term(norm_txt, txt, 'FPLX', id, id, 'assertion', 'famplex')
+            term = Term(norm_txt, txt, 'FPLX', id, id, 'curated', 'famplex')
         elif 'HGNC' in groundings:
             id = groundings['HGNC']
             term = Term(norm_txt, txt, 'HGNC', hgnc_client.get_hgnc_id(id), id,
-                        'assertion', 'famplex', '9606')
+                        'curated', 'famplex', '9606')
         elif 'UP' in groundings:
             db = 'UP'
             id = groundings['UP']
@@ -247,24 +259,25 @@ def generate_famplex_terms(ignore_mappings=False):
             # comes from and then add that organism info, otherwise
             # these groundings will be asserted even if the organism
             # doesn't match
-            term = Term(norm_txt, txt, db, id, name, 'assertion', 'famplex',
+            term = Term(norm_txt, txt, db, id, name, 'curated', 'famplex',
                         organism)
         elif 'CHEBI' in groundings:
             id = groundings['CHEBI']
             name = chebi_client.get_chebi_name_from_id(id[6:])
-            term = Term(norm_txt, txt, 'CHEBI', id, name, 'assertion',
+            term = Term(norm_txt, txt, 'CHEBI', id, name, 'curated',
                         'famplex')
         elif 'GO' in groundings:
             id = groundings['GO']
             term = Term(norm_txt, txt, 'GO', id,
-                        go_client.get_go_label(id), 'assertion', 'famplex')
+                        go_client.get_go_label(id), 'curated', 'famplex')
         elif 'MESH' in groundings:
             id = groundings['MESH']
             mesh_mapping = mesh_mappings.get(id)
             db, db_id, name = mesh_mapping if (mesh_mapping
                                                and not ignore_mappings) else \
                 ('MESH', id, mesh_client.get_mesh_name(id))
-            term = Term(norm_txt, txt, db, db_id, name, 'assertion', 'famplex')
+            term = Term(norm_txt, txt, db, db_id, name, 'curated', 'famplex',
+                        None, 'MESH', groundings['MESH'])
         else:
             # TODO: handle HMDB, PUBCHEM, CHEMBL
             continue
@@ -286,68 +299,110 @@ def generate_uniprot_terms(download=False, organisms=None):
         with open(path, 'w') as fh:
             fh.write(res.text)
     terms = []
+    logger.info('Loading %s' % path)
     for row in read_csv(path, delimiter='\t', header=True):
-        up_id = row['Entry']
-        organism = row['Organism ID']
-        protein_names = parse_uniprot_synonyms(row['Protein names'])
-        primary_gene_name = row['Gene names  (primary )'].strip()
-        if primary_gene_name == ';':
-            primary_gene_name = None
-        gene_synonyms_str = row['Gene names  (synonym )'].strip()
-        if gene_synonyms_str == ';':
-            gene_synonyms_str = None
-        # We generally use the gene name as the standard name
-        # except when there are multiple gene names (separated by
-        # semi-colons) in which case we take the first protein name.
-        if not primary_gene_name or ';' in primary_gene_name:
-            standard_name = protein_names[0]
-        else:
-            standard_name = primary_gene_name
-        # We skip a small number of not critical entries that don't have
-        # standard names
-        if not standard_name:
-            continue
-        ns = 'UP'
-        id = up_id
-        hgnc_id = uniprot_client.get_hgnc_id(up_id)
-        if hgnc_id:
-            ns = 'HGNC'
-            id = hgnc_id
-            standard_name = hgnc_client.get_hgnc_name(hgnc_id)
-        for name in protein_names:
-            # Skip names that are EC codes
-            if name.startswith('EC '):
-                continue
-            if name == standard_name:
-                continue
-            term = Term(normalize(name), name, ns, id,
-                        standard_name, 'synonym', 'uniprot',
-                        organism)
-            terms.append(term)
-        term = Term(normalize(standard_name), standard_name,
-                    ns, id, standard_name, 'name', 'uniprot',
-                    organism)
-        terms.append(term)
-        if gene_synonyms_str:
-            # This is to deal with all the variations in which
-            # synonyms are listed, including degenerate strings
-            # like "; ;"
-            for synonym_group in gene_synonyms_str.split('; '):
-                for synonym in synonym_group.split(' '):
-                    if not synonym or synonym == ';':
-                        continue
-                    term = Term(normalize(synonym), synonym,
-                                ns, id, standard_name, 'synonym', 'uniprot',
-                                organism)
-                    terms.append(term)
+        terms += get_terms_from_uniprot_row(row)
 
     return terms
 
 
+def get_terms_from_uniprot_row(row):
+    terms = []
+    up_id = row['Entry']
+    organism = row['Organism ID']
+
+    # As of 3/2/2022 there is an error in UniProt data that we need to manually
+    # patch here
+    if up_id == 'Q2QKR2':
+        row['Protein names'] = row['Protein names'][:-1]
+    protein_names = parse_uniprot_synonyms(row['Protein names'])
+
+    # These two lists are aligned and each separated by "; " if there
+    # are multiple genes. If there are no genes listed, we simply have
+    # an empty string here. If there are multiple genes but one doesn't
+    # have a name given, a corresponding "; " placeholder is still there.
+    # Consequently, we encounter cases like
+    # P34539»·; »·; »·
+    # where there are two genes but neither of them have names or
+    # synonyms listed.
+    primary_gene_names = row['Gene names  (primary )'].split('; ')
+    gene_synonyms = row['Gene names  (synonym )'].split('; ')
+
+    multi_gene = len(primary_gene_names) > 1
+
+    # We generally use the gene name as the standard name
+    # except when there are multiple gene names or the gene name is missing
+    if not primary_gene_names or multi_gene:
+        standard_name = protein_names[0]
+    else:
+        standard_name = primary_gene_names[0]
+    # We skip a small number of non-critical entries that don't have
+    # standard names
+    if not standard_name:
+        return []
+
+    # By default, we use the UniProt namespace and ID
+    ns = 'UP'
+    id = up_id
+    # For human genes, we resolve redundancies by mapping to HGNC
+    hgnc_id = uniprot_client.get_hgnc_id(up_id)
+    # We only map to HGNC if there is a single gene corresponding to
+    # this protein. Otherwise, we keep the UniProt ID.
+    if hgnc_id and not multi_gene:
+        ns = 'HGNC'
+        id = hgnc_id
+        standard_name = hgnc_client.get_hgnc_name(hgnc_id)
+
+    # We add all of the protein names as synonyms
+    for name in protein_names:
+        # Skip names that are EC codes
+        if name.startswith('EC '):
+            continue
+        if name == standard_name:
+            continue
+        term = Term(normalize(name), name, ns, id,
+                    standard_name, 'synonym', 'uniprot',
+                    organism, None if ns == 'UP' else 'UP',
+                    None if ns == 'UP' else up_id)
+        terms.append(term)
+
+    # We add the standard name (usually the gene name)
+    term = Term(normalize(standard_name), standard_name,
+                ns, id, standard_name, 'name', 'uniprot',
+                organism, None if ns == 'UP' else 'UP',
+                None if ns == 'UP' else up_id)
+    terms.append(term)
+
+    # If we have gene synonyms we include them according to the following logic.
+    # For human genes we do not include synonyms if there are multiple genes
+    # corresponding to this protein. For non-human genes, since we don't
+    # have a separate gene resource, we do include gene names and synonyms
+    # even if there are multiple genes corresponding to the protein.
+    if not ((organism == '9606') and multi_gene):
+        for gene_name, gene_synonyms_str in zip(primary_gene_names,
+                                                gene_synonyms):
+            all_synonyms = gene_synonyms_str.split(' ')
+            # We can skip a gene name if we used it as standard name
+            if gene_name and (gene_name != standard_name):
+                all_synonyms.append(gene_name)
+            for synonym in all_synonyms:
+                if not synonym:
+                    continue
+                term = Term(normalize(synonym), synonym,
+                            ns, id, standard_name, 'synonym', 'uniprot',
+                            organism, None if ns == 'UP' else 'UP',
+                            None if ns == 'UP' else up_id)
+                terms.append(term)
+    return terms
+
+
+ec_code_pattern = re.compile(r'EC [\d\.-]+')
+
+
 def parse_uniprot_synonyms(synonyms_str):
     synonyms_str = re.sub(r'\[Includes: ([^]])+\]',
                           '', synonyms_str).strip()
-    synonyms_str = re.sub(r'\[Cleaved into: ([^]])+\]',
+    synonyms_str = re.sub(r'\[Cleaved into: ([^]])+\]( \(Fragments\))?',
                           '', synonyms_str).strip()
 
     def find_block_from_right(s):
@@ -368,44 +423,54 @@ def parse_uniprot_synonyms(synonyms_str):
 
     syns = []
     while True:
+        # If the string is empty at this point, we return with the
+        # synonyms so far
         if not synonyms_str:
             return syns
+        # If the string doesn't end with a parenthesis, that means it's
+        # the first synonym which isn't in parentheses, and so we prepend
+        # it to the list of synonyms and return
         if not synonyms_str.endswith(')'):
             return [synonyms_str] + syns
 
         syn = find_block_from_right(synonyms_str)
-        syns = [syn] + syns
-        synonyms_str = synonyms_str[:-len(syn)-3]
+        # This is where we remove the synonym we just processed plus
+        # the space and opening/closing parens (3 characters) to get to the
+        # next synonym
+        to_remove = len(syn) + 3
+        # One corner case is when there happens to be a parenthesis in the
+        # first synonym in the list like "X(0) (...) (...)"
+        # In this case the entire synonyms_str is to be used and we can
+        # prepend and return
+        if synonyms_str[-to_remove] != ' ':
+            return [synonyms_str] + syns
+
+        # EC codes are not valid synonyms bot otherwise we prepend
+        # the synonym to the list of synonyms
+        if not ec_code_pattern.match(syn):
+            syns = [syn] + syns
+        # We now remove the processed suffix
+        synonyms_str = synonyms_str[:-to_remove]
 
 
 def generate_adeft_terms():
     from adeft import available_shortforms
     from adeft.disambiguate import load_disambiguator
+    from indra.ontology.standardize import get_standard_name
     all_term_args = set()
     for shortform in available_shortforms:
         da = load_disambiguator(shortform)
-        for grounding in da.names.keys():
+        for grounding, name in da.names.items():
             if grounding == 'ungrounded' or ':' not in grounding:
                 continue
             db_ns, db_id = grounding.split(':', maxsplit=1)
-            if db_ns == 'HGNC':
-                standard_name = hgnc_client.get_hgnc_name(db_id)
-            elif db_ns == 'GO':
-                standard_name = go_client.get_go_label(db_id)
-            elif db_ns == 'MESH':
-                standard_name = mesh_client.get_mesh_name(db_id)
-            elif db_ns == 'CHEBI':
-                standard_name = chebi_client.get_chebi_name_from_id(db_id)
-            elif db_ns == 'FPLX':
-                standard_name = db_id
-            elif db_ns == 'UP':
-                standard_name = uniprot_client.get_gene_name(db_id)
-            else:
-                logger.warning('Unknown grounding namespace from Adeft: %s' %
-                               db_ns)
-                continue
+            # Here we do a name standardization via INDRA just in case
+            # there is a discrepancy
+            indra_standard_name = get_standard_name({db_ns: db_id})
+            if indra_standard_name:
+                name = indra_standard_name
             term_args = (normalize(shortform), shortform, db_ns, db_id,
-                         standard_name, 'synonym', 'adeft')
+                         name, 'synonym', 'adeft')
             all_term_args.add(term_args)
     terms = [Term(*term_args) for term_args in sorted(list(all_term_args),
                                                       key=lambda x: x[0])]
@@ -417,7 +482,10 @@ def generate_doid_terms(ignore_mappings=False):
 
 
 def generate_efo_terms(ignore_mappings=False):
-    return _generate_obo_terms('efo', ignore_mappings)
+    terms = _generate_obo_terms('efo', ignore_mappings)
+    # We remove BFO terms since they are too generic to be useful
+    terms = [t for t in terms if not t.id.startswith('BFO:')]
+    return terms
 
 
 def generate_hp_terms(ignore_mappings=False):
@@ -477,6 +545,8 @@ def terms_from_obo_json_entry(entry, prefix, ignore_mappings=False,
         entry_name=name,
         status='name',
         source=prefix,
+        source_db=prefix.upper() if db != prefix.upper() else None,
+        source_id=entry['id'] if db != prefix.upper() else None,
     )
     terms.append(name_term)
 
@@ -505,6 +575,8 @@ def terms_from_obo_json_entry(entry, prefix, ignore_mappings=False,
             entry_name=name,
             status='synonym',
             source=prefix,
+            source_db=prefix.upper() if db != prefix.upper() else None,
+            source_id=entry['id'] if db != prefix.upper() else None,
         )
         terms.append(synonym_term)
     return terms
@@ -542,7 +614,7 @@ mesh_mappings, mesh_mappings_reverse = _make_mesh_mappings()
 def filter_out_duplicates(terms):
     logger.info('Filtering %d terms for uniqueness...' % len(terms))
     term_key = lambda term: (term.db, term.id, term.text)
-    statuses = {'assertion': 1, 'name': 2, 'synonym': 3, 'previous': 4}
+    statuses = {'curated': 1, 'name': 2, 'synonym': 3, 'former_name': 4}
     new_terms = []
     for _, terms in itertools.groupby(sorted(terms, key=lambda x: term_key(x)),
                                       key=lambda x: term_key(x)):
@@ -594,7 +666,12 @@ def main():
     terms = get_all_terms()
     from .resources import GROUNDING_TERMS_PATH as fname
     logger.info('Dumping into %s' % fname)
-    write_unicode_csv(fname, [t.to_list() for t in terms], delimiter='\t')
+    header = ['norm_text', 'text', 'db', 'id', 'entry_name', 'status',
+              'source', 'organism', 'source_db', 'source_id']
+    with gzip.open(fname, 'wt', encoding='utf-8') as fh:
+        writer = csv.writer(fh, delimiter='\t')
+        writer.writerow(header)
+        writer.writerows(t.to_list() for t in terms)
 
 
 if __name__ == '__main__':
