@@ -1,5 +1,7 @@
 """Module containing various string processing functions used for grounding."""
 import regex as re
+import unidecode
+
 from .greek_alphabet import greek_alphabet, greek_to_latin
 
 
@@ -79,6 +81,7 @@ def normalize(s):
     """
     s = replace_whitespace(s)
     s = remove_dashes(s)
+    s = replace_unicode(s)
     s = s.lower()
     return s
 
@@ -122,6 +125,17 @@ def replace_greek_spelled_out(s):
     return s
 
 
+def replace_unicode(s):
+    """Replace unicode with ASCII equivalent, except Greek letters.
+
+    Greek letters are handled separately and aren't replaced in this context.
+    """
+    if unidecode.unidecode(s) == s:
+        return s
+    return ''.join(unidecode.unidecode(c) if c not in greek_alphabet else c
+                   for c in s)
+
+
 def get_capitalization_pattern(word, beginning_of_sentence=False):
     """Return the type of capitalization for the string.
 
@@ -203,3 +217,39 @@ def depluralize(word):
         return word[:-1], 'plural_s'
     # Note: there don't seem to be any compelling examples of -f or -fe -> ves
     # so it is not implemented
+
+
+def replace_roman_arabic(s):
+    match = roman_arabic_prefilter.match(s)
+    if not match:
+        return s
+    else:
+        pattern = roman_arabic_patterns.get(match.groups()[0].upper())
+        return pattern[0].sub(pattern[1], s) if pattern else s
+
+
+def _make_roman_arabic_patterns():
+    roman_arabic = {
+        'I': '1',
+        'II': '2',
+        'III': '3',
+        'IV': '4',
+        'V': '5',
+        'VI': '6',
+        'VII': '7',
+        'VIII': '8',
+        'IX': '9',
+        'X': '10'
+    }
+
+    roman_arabic_patterns = {}
+    for r, a in roman_arabic.items():
+        for a, b in [(r, a), (a, r)]:
+            roman_arabic_patterns[a] = (re.compile(r'^(.*[- ])(%s)$' % a,
+                                                  re.IGNORECASE),
+                                        r'\g<1>%s' % b)
+    return roman_arabic_patterns
+
+
+roman_arabic_patterns = _make_roman_arabic_patterns()
+roman_arabic_prefilter = re.compile(r'^.*[- ](\d+|[IXV]+)$', re.IGNORECASE)
